@@ -1,5 +1,7 @@
 use super::{AsyncPushSender, ConnectionLike, Runtime, SharedHandleContainer, TaskHandle};
 use crate::aio::{check_resp3, setup_connection};
+#[cfg(feature = "cache-aio")]
+use crate::caching::{CacheManager, CacheStatistics, PrepareCacheResult};
 use crate::cmd::Cmd;
 #[cfg(any(feature = "tokio-comp", feature = "async-std-comp"))]
 use crate::parser::ValueCodec;
@@ -93,6 +95,19 @@ impl Debug for Pipeline {
     }
 }
 
+#[cfg(feature = "cache-aio")]
+pin_project! {
+    struct PipelineSink<T> {
+        #[pin]
+        sink_stream: T,
+        in_flight: VecDeque<InFlight>,
+        error: Option<RedisError>,
+        push_sender: Option<Arc<dyn AsyncPushSender>>,
+        cache_manager: Option<CacheManager>,
+    }
+}
+
+#[cfg(not(feature = "cache-aio"))]
 pin_project! {
     struct PipelineSink<T> {
         #[pin]
@@ -110,20 +125,18 @@ fn send_push(push_sender: &Option<Arc<dyn AsyncPushSender>>, info: PushInfo) {
 }
 
 pub(crate) fn send_disconnect(push_sender: &Option<Arc<dyn AsyncPushSender>>) {
-    send_push(
-        push_sender,
-        PushInfo {
-            kind: crate::PushKind::Disconnection,
-            data: vec![],
-        },
-    );
+    send_push(push_sender, PushInfo::disconnect());
 }
 
 impl<T> PipelineSink<T>
 where
     T: Stream<Item = RedisResult<Value>> + 'static,
 {
-    fn new(sink_stream: T, push_sender: Option<Arc<dyn AsyncPushSender>>) -> Self
+    fn new(
+        sink_stream: T,
+        push_sender: Option<Arc<dyn AsyncPushSender>>,
+        #[cfg(feature = "cache-aio")] cache_manager: Option<CacheManager>,
+    ) -> Self
     where
         T: Sink<Vec<u8>, Error = RedisError> + Stream<Item = RedisResult<Value>> + 'static,
     {
@@ -132,6 +145,8 @@ where
             in_flight: VecDeque::new(),
             error: None,
             push_sender,
+            #[cfg(feature = "cache-aio")]
+            cache_manager,
         }
     }
 
@@ -140,24 +155,18 @@ where
         loop {
             let item = ready!(self.as_mut().project().sink_stream.poll_next(cx));
             let item = match item {
-                Some(result) => {
-                    if let Err(err) = &result {
-                        if err.is_unrecoverable_error() {
-                            let self_ = self.as_mut().project();
-                            send_disconnect(self_.push_sender);
-                        }
-                    }
-                    result
-                }
-                // The redis response stream is not going to produce any more items so we `Err`
-                // to break out of the `forward` combinator and stop handling requests
-                None => {
-                    let self_ = self.project();
-                    send_disconnect(self_.push_sender);
-                    return Poll::Ready(Err(()));
-                }
+                Some(result) => result,
+                // The redis response stream is not going to produce any more items so we simulate a disconnection error to break out of the loop.
+                None => Err(closed_connection_error()),
             };
+
+            let is_unrecoverable = item.as_ref().is_err_and(|err| err.is_unrecoverable_error());
             self.as_mut().send_result(item);
+            if is_unrecoverable {
+                let self_ = self.project();
+                send_disconnect(self_.push_sender);
+                return Poll::Ready(Err(()));
+            }
         }
     }
 
@@ -166,6 +175,10 @@ where
         let result = match result {
             // If this push message isn't a reply, we'll pass it as-is to the push manager and stop iterating
             Ok(Value::Push { kind, data }) if !kind.has_reply() => {
+                #[cfg(feature = "cache-aio")]
+                if let Some(cache_manager) = &self_.cache_manager {
+                    cache_manager.handle_push_value(&kind, &data);
+                }
                 send_push(self_.push_sender, PushInfo { kind, data });
 
                 return;
@@ -341,6 +354,7 @@ impl Pipeline {
     fn new<T>(
         sink_stream: T,
         push_sender: Option<Arc<dyn AsyncPushSender>>,
+        #[cfg(feature = "cache-aio")] cache_manager: Option<CacheManager>,
     ) -> (Self, impl Future<Output = ()>)
     where
         T: Sink<Vec<u8>, Error = RedisError> + Stream<Item = RedisResult<Value>> + 'static,
@@ -352,7 +366,12 @@ impl Pipeline {
         const BUFFER_SIZE: usize = 50;
         let (sender, mut receiver) = mpsc::channel(BUFFER_SIZE);
 
-        let sink = PipelineSink::new(sink_stream, push_sender);
+        let sink = PipelineSink::new(
+            sink_stream,
+            push_sender,
+            #[cfg(feature = "cache-aio")]
+            cache_manager,
+        );
         let f = stream::poll_fn(move |cx| receiver.poll_recv(cx))
             .map(Ok)
             .forward(sink)
@@ -360,15 +379,6 @@ impl Pipeline {
         (Pipeline { sender }, f)
     }
 
-    // `None` means that the stream was out of items causing that poll loop to shut down.
-    async fn send_single(
-        &mut self,
-        item: Vec<u8>,
-        timeout: Option<Duration>,
-    ) -> Result<Value, Option<RedisError>> {
-        self.send_recv(item, None, timeout).await
-    }
-
     async fn send_recv(
         &mut self,
         input: Vec<u8>,
@@ -376,29 +386,34 @@ impl Pipeline {
         // If `Some`, the value inside defines how the response should look like
         expectation: Option<PipelineResponseExpectation>,
         timeout: Option<Duration>,
-    ) -> Result<Value, Option<RedisError>> {
+    ) -> Result<Value, RedisError> {
         let (sender, receiver) = oneshot::channel();
 
-        self.sender
-            .send(PipelineMessage {
-                input,
-                expectation,
-                output: sender,
-            })
-            .await
-            .map_err(|_| None)?;
+        let request = async {
+            self.sender
+                .send(PipelineMessage {
+                    input,
+                    expectation,
+                    output: sender,
+                })
+                .await
+                .map_err(|_| None)?;
+
+            receiver.await
+            // The `sender` was dropped which likely means that the stream part
+            // failed for one reason or another
+            .map_err(|_| None)
+            .and_then(|res| res.map_err(Some))
+        };
 
         match timeout {
-            Some(timeout) => match Runtime::locate().timeout(timeout, receiver).await {
+            Some(timeout) => match Runtime::locate().timeout(timeout, request).await {
                 Ok(res) => res,
-                Err(elapsed) => Ok(Err(elapsed.into())),
+                Err(elapsed) => Err(Some(elapsed.into())),
             },
-            None => receiver.await,
+            None => request.await,
         }
-        // The `sender` was dropped which likely means that the stream part
-        // failed for one reason or another
-        .map_err(|_| None)
-        .and_then(|res| res.map_err(Some))
+        .map_err(|err| err.unwrap_or_else(closed_connection_error))
     }
 }
 
@@ -424,6 +439,8 @@ pub struct MultiplexedConnection {
     // This handle is only set for connection whose task was spawned by the crate, not for users who spawned their own
     // task.
     _task_handle: Option<SharedHandleContainer>,
+    #[cfg(feature = "cache-aio")]
+    pub(crate) cache_manager: Option<CacheManager>,
 }
 
 impl Debug for MultiplexedConnection {
@@ -463,8 +480,7 @@ impl MultiplexedConnection {
             stream,
             AsyncConnectionConfig {
                 response_timeout,
-                connection_timeout: None,
-                push_sender: None,
+                ..Default::default()
             },
         )
         .await
@@ -490,16 +506,53 @@ impl MultiplexedConnection {
                 "Can only pass push sender to a connection using RESP3"
             );
         }
-        let (pipeline, driver) = Pipeline::new(codec, config.push_sender);
+
+        #[cfg(feature = "cache-aio")]
+        let cache_config = config.cache.as_ref().map(|cache| match cache {
+            crate::client::Cache::Config(cache_config) => *cache_config,
+            #[cfg(feature = "connection-manager")]
+            crate::client::Cache::Manager(cache_manager) => cache_manager.cache_config,
+        });
+        #[cfg(feature = "cache-aio")]
+        let cache_manager_opt = config
+            .cache
+            .map(|cache| {
+                check_resp3!(
+                    connection_info.protocol,
+                    "Can only enable client side caching in a connection using RESP3"
+                );
+                match cache {
+                    crate::client::Cache::Config(cache_config) => {
+                        Ok(CacheManager::new(cache_config))
+                    }
+                    #[cfg(feature = "connection-manager")]
+                    crate::client::Cache::Manager(cache_manager) => Ok(cache_manager),
+                }
+            })
+            .transpose()?;
+
+        let (pipeline, driver) = Pipeline::new(
+            codec,
+            config.push_sender,
+            #[cfg(feature = "cache-aio")]
+            cache_manager_opt.clone(),
+        );
         let mut con = MultiplexedConnection {
             pipeline,
             db: connection_info.db,
             response_timeout: config.response_timeout,
             protocol: connection_info.protocol,
             _task_handle: None,
+            #[cfg(feature = "cache-aio")]
+            cache_manager: cache_manager_opt,
         };
         let driver = {
-            let auth = setup_connection(connection_info, &mut con);
+            let auth = setup_connection(
+                connection_info,
+                &mut con,
+                #[cfg(feature = "cache-aio")]
+                cache_config,
+            );
 
             futures_util::pin_mut!(auth);
 
@@ -533,10 +586,35 @@ impl MultiplexedConnection {
     /// Sends an already encoded (packed) command into the TCP socket and
     /// reads the single response from it.
     pub async fn send_packed_command(&mut self, cmd: &Cmd) -> RedisResult<Value> {
+        #[cfg(feature = "cache-aio")]
+        if let Some(cache_manager) = &self.cache_manager {
+            match cache_manager.get_cached_cmd(cmd) {
+                PrepareCacheResult::Cached(value) => return Ok(value),
+                PrepareCacheResult::NotCached(cacheable_command) => {
+                    let mut pipeline = crate::Pipeline::new();
+                    cacheable_command.pack_command(cache_manager, &mut pipeline);
+
+                    let result = self
+                        .pipeline
+                        .send_recv(
+                            pipeline.get_packed_pipeline(),
+                            Some(PipelineResponseExpectation {
+                                skipped_response_count: 0,
+                                expected_response_count: pipeline.commands.len(),
+                                is_transaction: false,
+                            }),
+                            self.response_timeout,
+                        )
+                        .await?;
+                    let replies: Vec<Value> = crate::types::from_owned_redis_value(result)?;
+                    return cacheable_command.resolve(cache_manager, replies.into_iter());
+                }
+                _ => (),
+            }
+        }
         self.pipeline
-            .send_single(cmd.get_packed_command(), self.response_timeout)
+            .send_recv(cmd.get_packed_command(), None, self.response_timeout)
             .await
-            .map_err(|err| err.unwrap_or_else(closed_connection_error))
     }
 
     /// Sends multiple already encoded (packed) command into the TCP socket
@@ -548,7 +626,26 @@ impl MultiplexedConnection {
         offset: usize,
         count: usize,
     ) -> RedisResult<Vec<Value>> {
-        let result = self
+        #[cfg(feature = "cache-aio")]
+        if let Some(cache_manager) = &self.cache_manager {
+            let (cacheable_pipeline, pipeline, (skipped_response_count, expected_response_count)) =
+                cache_manager.get_cached_pipeline(cmd);
+            let result = self
+                .pipeline
+                .send_recv(
+                    pipeline.get_packed_pipeline(),
+                    Some(PipelineResponseExpectation {
+                        skipped_response_count,
+                        expected_response_count,
+                        is_transaction: cacheable_pipeline.transaction_mode,
+                    }),
+                    self.response_timeout,
+                )
+                .await?;
+
+            return cacheable_pipeline.resolve(cache_manager, result);
+        }
+        let value = self
             .pipeline
             .send_recv(
                 cmd.get_packed_pipeline(),
@@ -559,15 +656,19 @@ impl MultiplexedConnection {
                 }),
                 self.response_timeout,
             )
-            .await
-            .map_err(|err| err.unwrap_or_else(closed_connection_error));
-
-        let value = result?;
+            .await?;
         match value {
             Value::Array(values) => Ok(values),
             _ => Ok(vec![value]),
         }
     }
+
+    /// Gets [`CacheStatistics`] for current connection if caching is enabled.
+    #[cfg(feature = "cache-aio")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "cache-aio")))]
+    pub fn get_cache_statistics(&self) -> Option<CacheStatistics> {
+        self.cache_manager.as_ref().map(|cm| cm.statistics())
+    }
 }
 
 impl ConnectionLike for MultiplexedConnection {
@@ -590,12 +691,21 @@ impl ConnectionLike for MultiplexedConnection {
 }
 
 impl MultiplexedConnection {
-    /// Subscribes to a new channel.
+    /// Subscribes to a new channel(s).    
     ///
     /// Updates from the sender will be sent on the push sender that was passed to the connection.
-    /// If the connection was configured without a push sender, the connection won't be able to pass messages back to the user..
+    /// If the connection was configured without a push sender, the connection won't be able to pass messages back to the user.
     ///
     /// This method is only available when the connection is using RESP3 protocol, and will return an error otherwise.
+    ///
+    /// # async fn func() -> redis::RedisResult<()> {
+    /// let client = redis::Client::open("redis://127.0.0.1/?protocol=resp3").unwrap();
+    /// let (tx, mut rx) = tokio::sync::mpsc::unbounded_channel();
+    /// let config = redis::AsyncConnectionConfig::new().set_push_sender(tx);
+    /// let mut con = client.get_multiplexed_async_connection_with_config(&config).await?;
+    /// con.subscribe(&["channel_1", "channel_2"]).await?;
+    /// # Ok(()) }
+    /// # }
     pub async fn subscribe(&mut self, channel_name: impl ToRedisArgs) -> RedisResult<()> {
         check_resp3!(self.protocol);
         let mut cmd = cmd("SUBSCRIBE");
@@ -604,9 +714,19 @@ impl MultiplexedConnection {
         Ok(())
     }
 
-    /// Unsubscribes from channel.
+    /// Unsubscribes from channel(s).
     ///
     /// This method is only available when the connection is using RESP3 protocol, and will return an error otherwise.
+    ///
+    /// # async fn func() -> redis::RedisResult<()> {
+    /// let client = redis::Client::open("redis://127.0.0.1/?protocol=resp3").unwrap();
+    /// let (tx, mut rx) = tokio::sync::mpsc::unbounded_channel();
+    /// let config = redis::AsyncConnectionConfig::new().set_push_sender(tx);
+    /// let mut con = client.get_multiplexed_async_connection_with_config(&config).await?;
+    /// con.subscribe(&["channel_1", "channel_2"]).await?;
+    /// con.unsubscribe(&["channel_1", "channel_2"]).await?;
+    /// # Ok(()) }
+    /// # }
     pub async fn unsubscribe(&mut self, channel_name: impl ToRedisArgs) -> RedisResult<()> {
         check_resp3!(self.protocol);
         let mut cmd = cmd("UNSUBSCRIBE");
@@ -615,12 +735,22 @@ impl MultiplexedConnection {
         Ok(())
     }
 
-    /// Subscribes to a new channel with pattern.
+    /// Subscribes to new channel(s) with pattern(s).
     ///
     /// Updates from the sender will be sent on the push sender that was passed to the connection.
-    /// If the connection was configured without a push sender, the connection won't be able to pass messages back to the user..
+    /// If the connection was configured without a push sender, the connection won't be able to pass messages back to the user.
     ///
     /// This method is only available when the connection is using RESP3 protocol, and will return an error otherwise.
+    ///
+    /// # async fn func() -> redis::RedisResult<()> {
+    /// let client = redis::Client::open("redis://127.0.0.1/?protocol=resp3").unwrap();
+    /// let (tx, mut rx) = tokio::sync::mpsc::unbounded_channel();
+    /// let config = redis::AsyncConnectionConfig::new().set_push_sender(tx);
+    /// let mut con = client.get_multiplexed_async_connection_with_config(&config).await?;
+    /// con.subscribe(&["channel_1", "channel_2"]).await?;
+    /// con.unsubscribe(&["channel_1", "channel_2"]).await?;
+    /// # Ok(()) }
+    /// # }
     pub async fn psubscribe(&mut self, channel_pattern: impl ToRedisArgs) -> RedisResult<()> {
         check_resp3!(self.protocol);
         let mut cmd = cmd("PSUBSCRIBE");
@@ -629,7 +759,7 @@ impl MultiplexedConnection {
         Ok(())
     }
 
-    /// Unsubscribes from channel pattern.
+    /// Unsubscribes from channel pattern(s).
     ///
     /// This method is only available when the connection is using RESP3 protocol, and will return an error otherwise.
     pub async fn punsubscribe(&mut self, channel_pattern: impl ToRedisArgs) -> RedisResult<()> {
