@@ -32,7 +32,7 @@ use crate::streams;
 use crate::acl;
 use crate::RedisConnectionInfo;
 
-#[cfg(feature = "cluster")]
+#[cfg(any(feature = "cluster", feature = "cache-aio"))]
 pub(crate) fn is_readonly_cmd(cmd: &[u8]) -> bool {
     matches!(
         cmd,
@@ -130,6 +130,8 @@ pub(crate) fn is_readonly_cmd(cmd: &[u8]) -> bool {
             | b"ZSCAN"
             | b"ZSCORE"
             | b"ZUNION"
+            | b"JSON.GET"
+            | b"JSON.MGET"
     )
 }
 
@@ -244,12 +246,12 @@ implement_commands! {
         cmd("PEXPIREAT").arg(key).arg(ts)
     }
 
-    /// Get the time to live for a key in seconds.
+    /// Get the absolute Unix expiration timestamp in seconds.
     fn expire_time<K: ToRedisArgs>(key: K) {
         cmd("EXPIRETIME").arg(key)
-        }
+    }
 
-    /// Get the time to live for a key in milliseconds.
+    /// Get the absolute Unix expiration timestamp in milliseconds.
     fn pexpire_time<K: ToRedisArgs>(key: K) {
         cmd("PEXPIRETIME").arg(key)
     }
@@ -259,12 +261,12 @@ implement_commands! {
         cmd("PERSIST").arg(key)
     }
 
-    /// Get the expiration time of a key.
+    /// Get the time to live for a key in seconds.
     fn ttl<K: ToRedisArgs>(key: K) {
         cmd("TTL").arg(key)
     }
 
-    /// Get the expiration time of a key in milliseconds.
+    /// Get the time to live for a key in milliseconds.
     fn pttl<K: ToRedisArgs>(key: K) {
         cmd("PTTL").arg(key)
     }
@@ -395,7 +397,7 @@ implement_commands! {
         cmd("HSETNX").arg(key).arg(field).arg(value)
     }
 
-    /// Sets a multiple fields in a hash.
+    /// Sets multiple fields in a hash.
     fn hset_multiple<K: ToRedisArgs, F: ToRedisArgs, V: ToRedisArgs>(key: K, items: &'a [(F, V)]) {
         cmd("HMSET").arg(key).arg(items)
     }
@@ -414,17 +416,17 @@ implement_commands! {
         cmd("HEXISTS").arg(key).arg(field)
     }
 
-    /// Get one or more fields TTL in seconds.
+    /// Get one or more fields' TTL in seconds.
     fn httl<K: ToRedisArgs, F: ToRedisArgs>(key: K, fields: F) {
         cmd("HTTL").arg(key).arg("FIELDS").arg(fields.num_of_args()).arg(fields)
     }
 
-    /// Get one or more fields TTL in milliseconds.
+    /// Get one or more fields' TTL in milliseconds.
     fn hpttl<K: ToRedisArgs, F: ToRedisArgs>(key: K, fields: F) {
         cmd("HPTTL").arg(key).arg("FIELDS").arg(fields.num_of_args()).arg(fields)
     }
 
-    /// Set one or more fields time to live in seconds.
+    /// Set one or more fields' time to live in seconds.
     fn hexpire<K: ToRedisArgs, F: ToRedisArgs>(key: K, seconds: i64, opt: ExpireOption, fields: F) {
        cmd("HEXPIRE").arg(key).arg(seconds).arg(opt).arg("FIELDS").arg(fields.num_of_args()).arg(fields)
     }
@@ -444,7 +446,7 @@ implement_commands! {
         cmd("HPERSIST").arg(key).arg("FIELDS").arg(fields.num_of_args()).arg(fields)
     }
 
-    /// Set one or more fields time to live in milliseconds.
+    /// Set one or more fields' time to live in milliseconds.
     fn hpexpire<K: ToRedisArgs, F: ToRedisArgs>(key: K, milliseconds: i64, opt: ExpireOption, fields: F) {
         cmd("HPEXPIRE").arg(key).arg(milliseconds).arg(opt).arg("FIELDS").arg(fields.num_of_args()).arg(fields)
     }
@@ -587,6 +589,16 @@ implement_commands! {
         cmd("LSET").arg(key).arg(index).arg(value)
     }
 
+    /// Sends a ping to the server
+    fn ping<>() {
+         &mut cmd("PING")
+    }
+
+    /// Sends a ping with a message to the server
+    fn ping_message<K: ToRedisArgs>(message: K) {
+         cmd("PING").arg(message)
+    }
+
     /// Removes and returns the up to `count` last elements of the list stored at key
     ///
     /// If `count` is not specified, then defaults to last element.
@@ -1022,6 +1034,11 @@ implement_commands! {
         cmd("PUBLISH").arg(channel).arg(message)
     }
 
+    /// Posts a message to the given sharded channel.
+    fn spublish<K: ToRedisArgs, E: ToRedisArgs>(channel: K, message: E) {
+        cmd("SPUBLISH").arg(channel).arg(message)
+    }
+
     // Object commands
 
     /// Returns the encoding of a key.
@@ -1049,6 +1066,16 @@ implement_commands! {
         cmd("CLIENT").arg("GETNAME")
     }
 
+    /// Returns the ID of the current connection.
+    fn client_id<>() {
+        cmd("CLIENT").arg("ID")
+    }
+
+    /// Command assigns a name to the current connection.
+    fn client_setname<K: ToRedisArgs>(connection_name: K) {
+        cmd("CLIENT").arg("SETNAME").arg(connection_name)
+    }
+
     // ACL commands
 
     /// When Redis is configured to use an ACL file (with the aclfile
@@ -2150,6 +2177,10 @@ implement_commands! {
     // script commands
 
     /// Adds a prepared script command to the pipeline.
+    ///
+    /// Note: unlike a call to [`invoke`](crate::ScriptInvocation::invoke), if the script isn't loaded during the pipeline operation,
+    /// it will not automatically be loaded and retried. The script can be loaded using the
+    /// [`load`](crate::ScriptInvocation::load) operation.
     #[cfg_attr(feature = "script", doc = r##"
 
 # Examples:
@@ -2161,6 +2192,7 @@ implement_commands! {
 let script = redis::Script::new(r"
     return tonumber(ARGV[1]) + tonumber(ARGV[2]);
 ");
+script.prepare_invoke().load(&mut con)?;
 let (a, b): (isize, isize) = redis::pipe()
     .invoke_script(script.arg(1).arg(2))
     .invoke_script(script.arg(2).arg(3))
@@ -2176,6 +2208,54 @@ assert_eq!(b, 5);
     fn invoke_script<>(invocation: &'a crate::ScriptInvocation<'a>) {
         &mut invocation.eval_cmd()
     }
+
+    // cleanup commands
+
+    /// Deletes all the keys of all databases
+    ///
+    /// Whether the flushing happens asynchronously or synchronously depends on the configuration
+    /// of your Redis server.
+    ///
+    /// To enforce a flush mode, use [`Commands::flushall_options`].
+    ///
+    /// ```text
+    /// FLUSHALL
+    /// ```
+    fn flushall<>() {
+        &mut cmd("FLUSHALL")
+    }
+
+    /// Deletes all the keys of all databases with options
+    ///
+    /// ```text
+    /// FLUSHALL [ASYNC|SYNC]
+    /// ```
+    fn flushall_options<>(options: &'a FlushAllOptions) {
+        cmd("FLUSHALL").arg(options)
+    }
+
+    /// Deletes all the keys of the current database
+    ///
+    /// Whether the flushing happens asynchronously or synchronously depends on the configuration
+    /// of your Redis server.
+    ///
+    /// To enforce a flush mode, use [`Commands::flushdb_options`].
+    ///
+    /// ```text
+    /// FLUSHDB
+    /// ```
+    fn flushdb<>() {
+        &mut cmd("FLUSHDB")
+    }
+
+    /// Deletes all the keys of the current database with options
+    ///
+    /// ```text
+    /// FLUSHDB [ASYNC|SYNC]
+    /// ```
+    fn flushdb_options<>(options: &'a FlushDbOptions) {
+        cmd("FLUSHDB").arg(options)
+    }
 }
 
 /// Allows pubsub callbacks to stop receiving messages.
@@ -2249,7 +2329,7 @@ pub trait PubSubCommands: Sized {
 impl<T> Commands for T where T: ConnectionLike {}
 
 #[cfg(feature = "aio")]
-impl<T> AsyncCommands for T where T: crate::aio::ConnectionLike + Send + Sized {}
+impl<T> AsyncCommands for T where T: crate::aio::ConnectionLike + Send + Sync + Sized {}
 
 impl PubSubCommands for Connection {
     fn subscribe<C, F, U>(&mut self, channels: C, mut func: F) -> RedisResult<U>
@@ -2560,6 +2640,48 @@ impl ToRedisArgs for SetOptions {
     }
 }
 
+/// Options for the [FLUSHALL](https://redis.io/commands/flushall) command
+///
+/// # Example
+/// ```rust,no_run
+/// use redis::{Commands, RedisResult, FlushAllOptions};
+/// fn flushall_sync(
+///     con: &mut redis::Connection,
+/// ) -> RedisResult<()> {
+///     let opts = FlushAllOptions{blocking: true};
+///     con.flushall_options(&opts)
+/// }
+/// ```
+#[derive(Clone, Copy, Default)]
+pub struct FlushAllOptions {
+    /// Blocking (`SYNC`) waits for completion, non-blocking (`ASYNC`) runs in the background
+    pub blocking: bool,
+}
+
+impl FlushAllOptions {
+    /// Set whether to run blocking (`SYNC`) or non-blocking (`ASYNC`) flush
+    pub fn blocking(mut self, blocking: bool) -> Self {
+        self.blocking = blocking;
+        self
+    }
+}
+
+impl ToRedisArgs for FlushAllOptions {
+    fn write_redis_args<W>(&self, out: &mut W)
+    where
+        W: ?Sized + RedisWrite,
+    {
+        if self.blocking {
+            out.write_arg(b"SYNC");
+        } else {
+            out.write_arg(b"ASYNC");
+        };
+    }
+}
+
+/// Options for the [FLUSHDB](https://redis.io/commands/flushdb) command
+pub type FlushDbOptions = FlushAllOptions;
+
 /// Creates HELLO command for RESP3 with RedisConnectionInfo
 pub fn resp3_hello(connection_info: &RedisConnectionInfo) -> Cmd {
     let mut hello_cmd = cmd("HELLO");
