@@ -5,12 +5,13 @@ mod support;
 #[cfg(test)]
 mod basic {
     use assert_approx_eq::assert_approx_eq;
-    use redis::{cmd, ProtocolVersion, PushInfo, RedisConnectionInfo, ScanOptions};
+    use redis::{cmd, Client, ProtocolVersion, PushInfo, RedisConnectionInfo, Role, ScanOptions};
     use redis::{
         Commands, ConnectionInfo, ConnectionLike, ControlFlow, ErrorKind, ExistenceCheck,
         ExpireOption, Expiry, PubSubCommands, PushKind, RedisResult, SetExpiry, SetOptions,
         ToRedisArgs, Value,
     };
+    use redis_test::utils::get_listener_on_free_port;
     use std::collections::{BTreeMap, BTreeSet};
     use std::collections::{HashMap, HashSet};
     use std::io::Read;
@@ -204,6 +205,17 @@ mod basic {
         assert_eq!(redis::cmd("INCR").arg("foo").query(&mut con), Ok(43usize));
     }
 
+    #[test]
+    fn test_ping() {
+        let ctx = TestContext::new();
+        let mut con = ctx.connection();
+
+        let res: String = con.ping_message("foobar").unwrap();
+        assert_eq!(&res, "foobar");
+        let res: String = con.ping().unwrap();
+        assert_eq!(&res, "PONG");
+    }
+
     #[test]
     fn test_getdel() {
         let ctx = TestContext::new();
@@ -757,6 +769,23 @@ mod basic {
         assert_eq!(k2, 45);
     }
 
+    #[test]
+    fn test_pipeline_len() {
+        let mut pl = redis::pipe();
+
+        pl.cmd("PING").cmd("SET").arg(1);
+        assert_eq!(pl.len(), 2);
+    }
+
+    #[test]
+    fn test_pipeline_is_empty() {
+        let mut pl = redis::pipe();
+
+        assert!(pl.is_empty());
+        pl.cmd("PING").cmd("SET").arg(1);
+        assert!(!pl.is_empty());
+    }
+
     #[test]
     fn test_real_transaction() {
         let ctx = TestContext::new();
@@ -915,6 +944,72 @@ mod basic {
         assert_eq!(msg.get_payload(), Ok("bar".to_string()));
     }
 
+    #[test]
+    fn test_pubsub_send_ping() {
+        use std::sync::{Arc, Barrier};
+        let ctx = TestContext::new();
+        let mut con = ctx.connection();
+
+        // Connection for subscriber api
+        let mut pubsub_con = ctx.connection();
+
+        // Barrier is used to make test thread wait to publish
+        // until after the pubsub thread has subscribed.
+        let publish_barrier = Arc::new(Barrier::new(2));
+        let pubsub_barrier = publish_barrier.clone();
+        // Barrier is used to make pubsub thread wait to ping
+        // until after the test thread has published.
+        let ping_barrier = Arc::new(Barrier::new(2));
+        let ping_barrier_clone = ping_barrier.clone();
+
+        let thread = spawn(move || {
+            let mut pubsub = pubsub_con.as_pubsub();
+            pubsub.subscribe("foo").unwrap();
+
+            let _ = pubsub_barrier.wait();
+            let _ = ping_barrier_clone.wait();
+
+            if ctx.protocol == ProtocolVersion::RESP3 {
+                let message: String = pubsub.ping().unwrap();
+                assert_eq!(message, "PONG");
+            } else {
+                let message: Vec<String> = pubsub.ping().unwrap();
+                assert_eq!(message, vec!["pong", ""]);
+            }
+
+            if ctx.protocol == ProtocolVersion::RESP3 {
+                let message: String = pubsub.ping_message("foobar").unwrap();
+                assert_eq!(message, "foobar");
+            } else {
+                let message: Vec<String> = pubsub.ping_message("foobar").unwrap();
+                assert_eq!(message, vec!["pong", "foobar"]);
+            }
+
+            let msg = pubsub.get_message().unwrap();
+            assert_eq!(msg.get_channel(), Ok("foo".to_string()));
+            assert_eq!(msg.get_payload(), Ok(42));
+
+            let msg = pubsub.get_message().unwrap();
+            assert_eq!(msg.get_channel(), Ok("foo".to_string()));
+            assert_eq!(msg.get_payload(), Ok(23));
+        });
+
+        let _ = publish_barrier.wait();
+
+        redis::cmd("PUBLISH")
+            .arg("foo")
+            .arg(42)
+            .exec(&mut con)
+            .unwrap();
+
+        // We can also call the command directly
+        assert_eq!(con.publish("foo", 23), Ok(1));
+
+        ping_barrier.wait();
+
+        thread.join().expect("Something went wrong");
+    }
+
     #[test]
     fn pub_sub_subscription_to_multiple_channels() {
         let ctx = TestContext::new();
@@ -2043,7 +2138,6 @@ mod basic {
     fn test_client_getname() {
         let ctx = TestContext::new();
         let mut con = ctx.connection();
-
         redis::cmd("CLIENT")
             .arg("SETNAME")
             .arg("connection-name")
@@ -2052,4 +2146,54 @@ mod basic {
         let res: String = con.client_getname().unwrap();
         assert_eq!(res, "connection-name");
     }
+
+    #[test]
+    fn test_client_id() {
+        let ctx = TestContext::new();
+        let mut con = ctx.connection();
+        let _num: i64 = con.client_id().unwrap();
+    }
+
+    #[test]
+    fn test_client_setname() {
+        let ctx = TestContext::new();
+        let mut con = ctx.connection();
+        assert_eq!(con.client_setname("connection-name"), Ok(()));
+        let res: String = redis::cmd("CLIENT").arg("GETNAME").query(&mut con).unwrap();
+        assert_eq!(res, "connection-name");
+    }
+
+    #[test]
+    fn test_role_primary() {
+        let ctx = TestContext::new();
+        let mut con = ctx.connection();
+        let ret = redis::cmd("ROLE").query::<Role>(&mut con).unwrap();
+        assert!(matches!(ret, Role::Primary { .. }));
+    }
+
+    #[test]
+    fn test_connection_timeout_on_busy_server() {
+        let ctx = TestContext::new();
+        let mut con = ctx.connection();
+        // we stop the server on another thread, in order to move forwardd while the server is stopped.
+        let handle = thread::spawn(move || {
+            cmd("DEBUG")
+                .arg("SLEEP")
+                .arg("0.1")
+                .query::<()>(&mut con)
+                .unwrap();
+        });
+
+        // we wait, to ensure that the debug sleep has started.
+        thread::sleep(Duration::from_millis(5));
+        // we set a DB, in order to force calling requests on the server.
+        let mut addr = ctx.server.connection_info().clone();
+        addr.redis.db = 1;
+        let client = Client::open(addr).unwrap();
+        let try_connect = client.get_connection_with_timeout(Duration::from_millis(2));
+
+        assert!(try_connect.is_err_and(|err| { err.is_timeout() }));
+
+        handle.join().unwrap();
+    }
 }
